#ifndef _EBOOK_H_
#define _EBOOK_H_

#include <inttypes.h>

// the persistent values that may reside in EEPROM
typedef struct
{
    uint8_t bright;
    uint8_t speed;
    uint8_t show_vers;
} eeprom_t;

typedef struct
{
    // the address that we read from.
    // to get it physical call prepare_addr()
    uint32_t addr;
 
    // number of the verse (as 8-bit)
    uint8_t nvers;

    // name of the book
    char name[24];

    // the chapter (as string)
    // points somewhere into name[]
    char *kapitel;

    eeprom_t eeprom;
 
} ebook_t;

extern ebook_t ebook;

// lives in EEPROM (EEMEM, i.e. section ".eeprom")
extern eeprom_t eeprom;


// Set ebook.name to book's name and ebook.kapitel to the chapter.
// ebook.addr must point just after the '_' that marks a book's start.
// Then set ebook.addr to the first letter of the book's first verse.
extern void read_book_name();

/*
  pre-decrement ebook.addr by 1 if dir < 0
  pre-increment ebook.addr by 1 if dir > 0
   
  Set address and ChipEnable (output latches of port expanders)
  to correspond with abook.addr. The ChipEnable lines are
  expanded from bits 20 and 21 of ebook.addr so that we have the
  following address spaces:
   
  0x000000 - 0x0fffff : EPROM #1 (CE1 = LOW, other CEn = HIGH)
  0x100000 - 0x1fffff : EPROM #2 (CE2 = LOW, other CEn = HIGH)
  0x200000 - 0x2fffff : EPROM #3 (CE3 = LOW, other CEn = HIGH)
  0x300000 - 0x3fffff : EPROM #4 (CE4 = LOW, other CEn = HIGH)
  
  To set the address bus lines to a specific address foo use
     ebook.addr = foo;
     prepare_addr (0);
*/
extern void prepare_addr (const char dir);

// Read one byte from the EPROMs and the actual address
// (the address that is at the output latches of the
// port expanders
extern uint8_t read_byte();

// seek the |dir|-th occurrence of b
// dir < 0: seek backward
// dir > 0: seek forward
// the return value is the number of '\n's that are encountered
// during the seek
extern uint8_t seek_byte (uint8_t b, char dir);

// seek a book and position ebook.addr to the first letter of the
// book's first verse.
// Set ebook.name to book's name and abook.kapitel to the chapter.
// dir =  0: this book
// dir = -1: previous book
// dir =  1: next book
extern void seek_buch (char dir);

// Seek a chapter
// dir =  0: this chapter
// dir = -1: previous chapter
// dir =  1: next chapter
// Set ebook.name and ebook.kapitel and set ebook.addr to the first letter
// of the chapter's first verse.
extern void seek_kapitel (char dir);


// Seek a verse
// dir =  0: this verse
// dir = -1: previous verse
// dir =  1: next verse
// Set ebook.addr to the first letter of the verse.
extern void seek_vers (char dir);

// initialize ebook
// - clear VFD and reset cursors
// - set VFD's brightness according to ebook.eeprom.bright
// - set ebook.seekmode = SEEK_BOOK
extern void ebook_init ();

#endif /* _EBOOK_H_ */
