package de.gjlay.tools;

import java.util.*;

/**
 * This class can be used for inspection and
 * classification of command line parameters as they
 * are passed to a <code>public static void main(String[])</code>
 * method.
 * <p>
 * <code>Arguments</code> destinguishes between three kinds
 * of parameters:
 * <dl>
 * <dt>Switches</dt>
 * <dd>
 * Switches are of the form
 * <br><code>-key</code><br>
 * A switch is said to be 'on' if it is present and
 * said to be 'off', otherwise.
 * </dd>
 * <dt>Options</dt>
 * <dd>
 * Options are of the form:
 * <br><code>-key value</code><br>
 * An option may be present more then once.
 * </dd>
 * <dt>Arguments</dt>
 * <dd>
 * An 'argument' is everything that is not a switch, not an option
 * and not the value of an option.
 * </dd>
 * </dl>
 * <h3>Example:</h3>
 * Suppose:
 * <br>
 * <code>
 * String[] sw = {"-sw1", "-sw2"};<br>
 * String[] op = {"-op1", "-op2"};<br>
 * Arguments args = new Arguments (sw, op);<br>
 * args.parse (params, true);<br>
 * </code>
 * then,  you will get (e.g.) with:<br>
 * <dl>
 * <dt><code>params = {"-sw2", "foo" "-op1", "-sw1", "bar"}</code></dt>
 * <dd>
 * <code>args.hasSwitch ("-sw2") = true;</code><br>
 * <code>args.hasSwitch ("-op1") = false;</code><br>
 * <code>args.getOption ("-op1") = "-sw1";</code><br>
 * <code>args.getArgument (0) = "foo";</code><br>
 * <code>args.getArgument (1) = "bar";</code><br>
 * <code>args.getArgumentCount () = 2;</code><br>
 * </dd>
 * <dt><code>params = {"-bar"}</code></dt>
 * <dd>
 * will throw an {@link ArgumentException} because "-bar"
 * is neither a switch nor an option and parsing is being
 * done strictly.
 * </dd>
 * </dl>
 * </code>
 * @author Georg-Johann Lay
 */

public class Arguments
	extends Object
{
	private HashSet switches;
	private Hashtable options;
	private Vector arguments;
	private HashSet allSwitches;
	private HashSet allOptions;

	private Arguments ()
		{
		}

	/**
	 * Construct a new Arguments object that recognizes
	 * switches from <code>switches</code> and
	 * options from <code>options</code>
	 * @param switches	the switches that may be set
	 * @param options	the options that may be given
	 * @see #getAllSwitches()
	 * @see #getAllOptions()
	 */
	public Arguments (String switches[], String options[])
		{
			this.switches = new HashSet();
			this.options = new Hashtable();
			this.arguments = new Vector();

			allSwitches = new HashSet();
			allOptions = new HashSet();

			if (null != switches)
				for (int i=0; i<switches.length; i++)
					allSwitches.add (switches[i]);

			if (null != options)
				for (int i=0; i<options.length; i++)
					allOptions.add (options[i]);
		}

	/**
	 * Parse arguments.
	 * @param params	an array of parameters like they are passed
	 *					to a <code>public static void main(String[])</code>
	 *					method.
	 * @param strict	the strict version expects every parameter (element
	 *					of <code>params</code>) that begins with
	 *					a '<code>-<code>'
	 *					(hyphen) to be a switch or an option. The only
	 *					exception to this is the string "<code>-</code>"
	 *					itself, that will be classified as argument.
	 */
	public void parse (String params[], boolean strict)
		throws ArgumentException
		{
			switches.clear();
			options.clear();
			arguments.clear();

			String unknown = null;
			
			if (null == params)
				return;
			
			for (int num = 0; num < params.length; num++)
			{
//				System.err.println ("ARG["+num+"]="+params[num]);
				if (allSwitches.contains (params[num]))
				{
					switches.add (params[num]);
					continue;
				}

				if (allOptions.contains (params[num]))
				{
					if (num+1 == params.length)
						throw new ArgumentException (
							"Option '"+params[num]+"' requires value.");
					if (!params[num+1].equals (""))
						addOption (params[num], params[num+1]);
					num++;
					continue;
				}

				if (strict)
				{
					if (params[num].startsWith ("-"))
						if (!params[num].equals ("-"))
							if (null == unknown)
								unknown = params[num];
				}

				if (!params[num].equals (""))
					arguments.add (params[num]);
			}

			if (null != unknown)
			{
				throw new ArgumentException (
					"Unknown option or switch '"+unknown+"'.");
			}
		}

	/**
	 * Is an option given?
	 * @param	option	the option to check for presence
	 * @return	<code>true</code> if this option was recognized
	 */
	public boolean hasOption (String option)
		{
			return (null != options.get (option));
		}

	/**
	 * Get the value of an option
	 * @param option	the option
	 * @return			the value that is associated
	 *					to option <code>option</code>, or <code>null</code>.
	 *					If the option was specified more than once,
	 *					then return the first value seen.
	 */
	public String getOption (String option)
		{
			if (!hasOption (option))
				return null;
			
			return (String) getOptions (option).get(0);
		}

	/**
	 * Get a list of all option values for the same option.
	 * @param option	the option
	 * @return			a list of strings of values that are associated
	 *					to option <code>option</code>, or <code>null</code>.
	 */
	public List getOptions (String option)
		{
			if (!hasOption (option))
				return null;
			
			return (List) options.get (option);
		}

	/**
	 * Add an option.
	 * @param key	the option to set
	 * @param value	the value for the option
	 */
	private void addOption (String key, String value)
		{
			List values = getOptions (key);

			if (null == values)
				values = new Vector();

			values.add (value);
			
			options.put (key, values);
		}

	/**
	 * Get all parameters that have been recognized as option.
	 * @return	a hashtable that contains all options and their
	 *			assigned values (both as strings)
	 */
	public Hashtable getOptions ()
		{
			return options;
		}
	
	/**
	 * Is a switch turned on?
	 * @param aSwitch	the switch
	 * @return			is the specified switch turned on (present)?
	 */
	public boolean hasSwitch (String aSwitch)
		{
			return  switches.contains (aSwitch);
		}

	/**
	 * Get all parameters that have been recognized as switch.
	 * @return		all switches as Collection of strings
	 */
	public Collection getSwitches ()
		{
			return switches;
		}

	/**
	 * Get the number of arguments.
	 * @return	the number of arguments given.
	 */
	public int getArgumentCount()
		{
			return arguments.size();
		}

	/**
	 * Get all parameters that have been recognized as argument.
	 * @return	a List of strings of all arguments in the order
	 *			they occur
	 */
	public List getArguments()
		{
			return arguments;
		}

	/**
	 * Get the argument at a specified position.
	 * @param index		the index
	 * @return			the argument at position <code>index</code>.
	 *					Numbering starts at 0.
	 */
	public String getArgument (int index)
		{
			return (String) arguments.elementAt (index);
		}

	/**
	 * Get all switches that may be set.
	 * @return	a {@link Set} of all switches as given in
	 *			{@link #Arguments(String[],String[])}.
	 */
	public Set getAllSwitches ()
		{
			return allSwitches;
		}
	
	/**
	 * Get all options that may be given.
	 * @return	a {@link Set} of all options as given in
	 *			{@link #Arguments(String[],String[])}.
	 */
	public Set getAllOptions ()
		{
			return allOptions;
		}
}

