#include <stdarg.h>
#include <string.h>
#include <div.h>

#include "vfd.h"
#include "vfd-put.h"

static void vfd_puts     (const char*);
static void vfd_puts_P   (const char*);// __attribute__ ((noinline));
static void vfd_putsint  (int16_t);
static void vfd_putuint  (uint16_t);
static void vfd_put_va   (const char *, va_list) __attribute__ ((noinline));

typedef struct
{
    // to track the number of printed characters
    uint8_t char_count;
    
    // when printing to a string with %+ and %@ (channel %2)
    char * put_at;

    // index into vfd_putters[] for actual character putter
    uint8_t putter_id;

    struct
    {
        // last output position, one of %0 or %1
        char pos;

        // position of %0
        char pos0;

        // position of %1
        char pos1;

        // top left, physical position where
        // put_adjust_cursor (0) will end
        char home;
    } cursor;
} vfd_put_t;

vfd_put_t vfd_put;

// dispatch to actual putter function and increment character count
static void putc (char c)
{
    vfd_putters[vfd_put.putter_id] (c);
    vfd_put.char_count++;
}

// compute cursors' positions with respect to the visible window. 
// set appropriate putter
// %0 if position is 0...58
// %1 if position is 0x40-8...0x40+39
static void put_adjust_cursor (char);

// compute cursors' positions and set VFD's cursor
// set appropriate putter
static void put_set_cursor (char);

void vfd_clear()
{
    vfd_put.cursor.pos = 0;
    vfd_put.cursor.pos0 = 0;
    vfd_put.cursor.home = 0;
    vfd_put.cursor.pos1 = VFD_LINE1_START;
    vfd_init(); 
    put_adjust_cursor (0);
}

// putter for line %0 and line %1
void vfd_putter_0 (char c)
{
    put_set_cursor (vfd_put.cursor.pos);
    vfd_putc (c);
    put_adjust_cursor (vfd_put.cursor.pos+1);
}

// putter for line %2
// write c to * vfd_put.put_at
void vfd_putter_at (char c)
{
    char *at = vfd_put.put_at;
    
    *at++ = c;
    *at   = '\0';
    vfd_put.put_at = at;
}

// array of character putters
// !!! PUTTER_CHANNELS MUST HAVE BEEN DEFINED ADEQUATELY
// !!! ARRAY MUST HAVE BEEN INITIALIZED WITH PROPER PUTTER FUNCTIONS
// !!! IF YOU USE CHANNEL %3 OR HIGHER
void (*vfd_putters[PUTTER_CHANNELS])(char) =
{
    vfd_putter_0,
    vfd_putter_0,
    vfd_putter_at
};

// Write a null terminated string from RAM
void vfd_puts (const char *s)
{
    while (1)
    {
        char c = *s++;

        if ('\0' == c)
            break;

        putc (c);
    }
}

// Write a null terminated string from FLASH
void vfd_puts_P (const char * s)
{
    while (1)
    {
        char c = pgm_read_byte (s++);

        if ('\0' == c)
            break;

        putc (c);
    }
}

// output a signed 16-bit integer
void vfd_putsint (int16_t zahl)
{
    if (zahl < 0)
    {
        putc ('-');
        zahl = -zahl;
    }
    
    vfd_putuint ((uint16_t) zahl);
}

// output an unsigned 16-bit integer
void vfd_putuint (uint16_t zahl)
{
    // 5 Ziffern (0...65535) und evtl. noch eine fhrende 0
    // +1 wegen String-Ende (wird im Startup auf 0 gesetzt)
    
    static char string[5+1];
    // s zeigt auf das Ende von string
    // string wird von hinten nach vorne gefllt
    char *s = string + 5;
    
    // qrem enthlt Quotient (quot) und Rest (rem) der Divisionen
    udiv_t qrem = {.quot = zahl};                  

    do
    {
        // Division mit Rest durch 10
        // quot: Ergebnis fr den nchsten Durchlauf
        // rem:  Rest ist die Ziffer im 10er-System
        qrem = udiv (qrem.quot, 10);
        
        // Ziffer in Zeichen wandeln und speichern
        *(--s) = '0' + qrem.rem;
    }    
    while (0 != qrem.quot);
    
    // Falls eine fhrende '0' gespeichert wurde: weg damit
    // ausser zahl war selbst schon 0
    if (s[0] == '0' && s[1] != '\0')
        s++;

    vfd_puts (s);
}

// Users' interface to this module, via MACRO vfd_put
// fmt is the format string that resides in the Flash
uint8_t vfd_put_PSTR (const char * fmt, ...)
{
    vfd_put.char_count = 0;
    
    va_list args;
    va_start (args, fmt);
    vfd_put_va (fmt, args);
    va_end (args);
    
    return vfd_put.char_count;
}

void put_adjust_cursor (char pos)
{
    if (pos >= VFD_LINE1_START - 8)
    {
        if (pos >= VFD_LINE1_START + VFD_LINE_SIZE)  pos -= VFD_LINE_SIZE;
        if (pos < VFD_LINE1_START)                   pos += VFD_LINE_SIZE;
        vfd_put.cursor.pos1 = pos;
        vfd_put.putter_id = 0;
    }
    else
    {
        if (pos >= VFD_LINE_SIZE)                    pos -= VFD_LINE_SIZE;
        if (pos < 0)                                 pos += VFD_LINE_SIZE;
        vfd_put.cursor.pos0 = pos;
        vfd_put.putter_id = 1;
    }
    
    vfd_put.cursor.pos = pos;
}

void put_set_cursor (char pos)
{
    put_adjust_cursor (pos);
    vfd_set_cursor (vfd_put.cursor.pos);
}

void vfd_put_va (const char * fmt, va_list args)
{
    put_adjust_cursor (vfd_put.cursor.pos);

    while (1)
    {
        char c = pgm_read_byte (fmt++);
        
        if ('\0' == c)
            return;

        if ('%' != c)
        {
            putc (c);
            continue;
        }

        // %-codes

        c = pgm_read_byte (fmt++);
            
        char home = vfd_put.cursor.home;
            
        if ('%' == c)  putc (c);
        else if ('c' == c)  putc         (va_arg (args, int));
        else if ('u' == c)  vfd_putuint  (va_arg (args, uint16_t));
        else if ('d' == c)  vfd_putsint  (va_arg (args, int16_t));
        else if ('s' == c)  vfd_puts     (va_arg (args, char*));
        else if ('S' == c)  vfd_puts_P   (va_arg (args, char*));
        else if ('p' == c)  put_adjust_cursor (home + va_arg (args, uint16_t));
        else if ('0' == c)  put_adjust_cursor (vfd_put.cursor.pos0);
        else if ('1' == c)  put_adjust_cursor (vfd_put.cursor.pos1);
        else if ('A' == c)  put_adjust_cursor (home + 0);
        else if ('B' == c)  put_adjust_cursor (home + VFD_VISIBLE_WIDTH);
        else if ('C' == c)  put_adjust_cursor (home + VFD_LINE1_START);
        else if ('D' == c)  put_adjust_cursor (home + VFD_LINE1_START + VFD_VISIBLE_WIDTH);
        else if ('^' == c)  vfd_clear();
        else if ('*' == c)
        {
            uint8_t count = va_arg (args, uint16_t);
            char c = pgm_read_byte (fmt++);
                
            while (count--)
                putc (c);
        }
        else if ('+' == c || '@' == c)    
        {
            char * at = va_arg (args, char*);
                
            if ('+' == c)
                at += strlen (at);
            
            vfd_put.put_at = at;
            vfd_put.putter_id = 2;
        }
        else if ('r' == c)
        {
            // we move the cursor just one place at a time so that we don't
            // accidently change to top/bottom line
            char r = va_arg (args, int);
            char dir = (r < 0) ? -1 : 1;

            while (r)
            {
                put_adjust_cursor (vfd_put.cursor.pos + dir);
                r -= dir;
            }
        }
        else if ('<' == c) 
        {
            // left shift moves home one to the right
            if (++home >= VFD_LINE_SIZE)
                home -= VFD_LINE_SIZE;
            vfd_put.cursor.home = home;

            // shift display 1 position to the left
            vfd_send0 ((1 << VFD_SHFIT_SELECT) | (1 << 3)); 

            // adjust cursors
            uint8_t pos = vfd_put.cursor.pos;    
            put_set_cursor (1 + vfd_put.cursor.pos0);    
            put_set_cursor (1 + vfd_put.cursor.pos1);    
            put_set_cursor (1 + pos);    
        }
        else if (c >= '2' && c <= '9' && c < '0' + PUTTER_CHANNELS)
            vfd_put.putter_id = (uint8_t) c-'0';
        else
        {
            // be eloquent if there is an unsupported %-code
            putc ('%');
            putc (c);
        }
    } // fmt while loop
}
