#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "vfd.h"
#include "timer.h"

#include "ports.h"

#include "vfd-symbols.c"

vfd_t vfd =
{
    .cursor_on = 1,
    .cursor_blink = 1,
    .bright = 1,
    .address_mode =1,
    .shift_mode = 0
};

#define VFD_DELAY(x) _delay_loop_1 (x)

static void vfd_write_symbol (vfd_symbol_t*);
static uint8_t vfd_receive_byte_or_nibble ();
static void vfd_send_byte_or_nibble (uint8_t b);
static char vfd_replace_uml (char);

// Wait until VFD is ready
void vfd_wait_busy()
{
    while ((1 << VFD_BUSY) & vfd_receive ());
}

// Set VFD with desired settings for blink mode, brightness, etc.
// Clear the VFD and set cursor home
void vfd_init ()
{
    uint8_t data;
 
    vfd_wait_busy();
    vfd_send0 (1 << VFD_DISPLAY_CLEAR);
 
    vfd_wait_busy();
    data = 1 << VFD_ON_OFF_CTRL;
    if (vfd.cursor_blink) data |= 1<<0;
    if (vfd.cursor_on)    data |= 1<<1;
    if (vfd.bright)       data |= 1<<2;
    vfd_send0 (data);
 
    vfd_wait_busy();
 
#ifdef VFD_INTERFACE_4BIT
    vfd_send0 (VFD_DATA4);
#endif /* VFD_INTERFACE_4BIT */

#ifdef VFD_INTERFACE_8BIT_PARALLEL
    vfd_send0 (VFD_SET_DATA8);
#endif /* VFD_INTERFACE_8BIT_PARALLEL */
    vfd_send1 (4 - vfd.bright);

    vfd_wait_busy();
 
    data = 1 << VFD_ENTRY_MODE_SET;
    if (vfd.shift_mode)     data |= 1<<0;
    if (vfd.address_mode)   data |= 1<<1;
 
    vfd_send0 (data);
    vfd_wait_busy();
}

// Send a byte with RS = 0 (command)
void vfd_send0 (uint8_t b)
{
    CLR (PORT_RS);
#ifdef VFD_INTERFACE_4BIT
    vfd_send_byte_or_nibble (b >> 4);
#endif /* VFD_INTERFACE_4BIT */
    vfd_send_byte_or_nibble (b);
}

// Send a byte with RS = 1 (data)
void vfd_send1 (uint8_t b)
{
    SET (PORT_RS);
#ifdef VFD_INTERFACE_4BIT
    vfd_send_byte_or_nibble (b>>4);
#endif /* VFD_INTERFACE_4BIT */
    vfd_send_byte_or_nibble (b);
}

// Receive 1 byte from the VFD
uint8_t vfd_receive ()
{
    uint8_t result;
 
    CLR (PORT_RS);
    CLR (PORT_E);
 
    VFD_IN_DB;
    VFD_CLR_DB;

#ifdef VFD_INTERFACE_4BIT
    result = vfd_receive_byte_or_nibble () << 4;
    VFD_DELAY (10);
    result |= vfd_receive_byte_or_nibble () & 0xf;
#endif /* VFD_INTERFACE_4BIT */

#ifdef VFD_INTERFACE_8BIT_PARALLEL
    result = vfd_receive_byte_or_nibble();
#endif /* VFD_INTERFACE_8BIT_PARALLEL */

    return result;
}

// Read 1 byte   from VFD (VFD_INTERFACE_8BIT defined)
// Read 1 nibble from VFD (VFD_INTERFACE_4BIT defined)
uint8_t vfd_receive_byte_or_nibble ()
{
    uint8_t bibble = 0;
 
    SET (PORT_RW);
    SET (PORT_E);
    VFD_DELAY (10);
 
#ifdef VFD_INTERFACE_4BIT
    if (IS_SET (PORT_DB0))   bibble |= (1<<0);
    if (IS_SET (PORT_DB1))   bibble |= (1<<1);
    if (IS_SET (PORT_DB2))   bibble |= (1<<2);
    if (IS_SET (PORT_DB3))   bibble |= (1<<3);
#endif /* VFD_INTERFACE_4BIT */

#ifdef VFD_INTERFACE_8BIT_PARALLEL
    bibble = N_TO_PIN (PORT_VFD_ID);
#endif /* VFD_INTERFACE_8BIT_PARALLEL */

    CLR (PORT_E);
    CLR (PORT_RW);

    return bibble;
}

// Send 1 byte   to VFD (VFD_INTERFACE_8BIT defined)
// Send 1 nibble to VFD (VFD_INTERFACE_4BIT defined)
void vfd_send_byte_or_nibble (uint8_t b)
{
    CLR (PORT_E);
 
    VFD_OUT_DB;
    CLR (PORT_RW);
 
#ifdef VFD_INTERFACE_4BIT
    VFD_CLR_DB;
    if (b & (1<<0))  SET (PORT_DB0);
    if (b & (1<<1))  SET (PORT_DB1);
    if (b & (1<<2))  SET (PORT_DB2);
    if (b & (1<<3))  SET (PORT_DB3);
#endif /* VFD_INTERFACE_4BIT */

#ifdef VFD_INTERFACE_8BIT_PARALLEL
    N_TO_PORT (PORT_VFD_ID) = b;
#endif /* VFD_INTERFACE_8BIT_PARALLEL */

    SET (PORT_E);
    VFD_DELAY (10);
    CLR (PORT_E);

    VFD_IN_DB;
}

// Move cursor to position
// top    line: 0..39
// bottom line: 0x40..0x40+39
void vfd_set_cursor (uint8_t pos)
{
    vfd_wait_busy();
    vfd_send0 (0x80 | pos);
    vfd_wait_busy();
}

// Write one character to the VFD
void vfd_putc (char c)
{
    vfd_send1 (vfd_replace_uml (c));
}

// Write extra character font to character generator RAM
void vfd_write_symbol (vfd_symbol_t *sym)
{
    uint8_t i;
    uint8_t id = pgm_read_byte (& sym->id);
    uint8_t *pdata = (uint8_t*) & sym->dots;
 
    id = (uint8_t) (id << 3) | (1 << VFD_CG_RAM);
 
    for (i=0; i < 8; i++)
    {
        vfd_wait_busy();
        vfd_send0 (id+i);
        uint8_t data = pgm_read_byte (pdata++);
  
        vfd_send1 (data);
    }
}

// Reset VFD by issuing a RESET signal
// Write special characters to VFD's character generator RAM
void vfd_reset ()
{
    CLR (PORT_RESET);
    wait_10ms (10);
    SET (PORT_RESET);
    wait_10ms (10);

    vfd_wait_busy();
 
    CLR (PORT_RS);

#ifdef VFD_INTERFACE_4BIT
    vfd_send_byte_or_nibble (VFD_SET_DATA4 >> 4);
#endif /* VFD_INTERFACE_4BIT */

#ifdef VFD_INTERFACE_8BIT_PARALLEL
    vfd_send_byte_or_nibble (VFD_SET_DATA8);
#endif /* VFD_INTERFACE_8BIT_PARALLEL */

    // bright = 25%
    vfd_send1 (3);
    vfd_wait_busy();

    // Write extra characters to character gernerator area
    vfd_write_symbol ((vfd_symbol_t*) &sym_szlig);
}

// Map umlauts and ligatures to their code number
// (slot number in character generator RAM)
char vfd_replace_uml (char c)
{
    static const char UML_MAP[][2] PROGMEM =
        {
            { '\n', 0x9d},
            { 'g', 0xe7},
            { '', 0x08},
            { '', 0xe1},
            { '', 0x87},
            { '', 0x8b},
            { '', 0x80},
            { '', 0x86},
            { '', 0x8a},
            { 0  , 0}
        };
    
    char c0;
    char *map = (char*) UML_MAP;

    do
    {
        c0 = pgm_read_byte (map);
  
        if (c0 == c)
        {
            c = pgm_read_byte (map+1);
            break;
        }
        map += 2;
    } while (c0);
 
    return c;
}
